#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove


"""klove: a Python module for studying resonators on thin elastic plates.
"""


from .__about__ import __author__, __description__, __version__

try:
    import numdiff
    from numdiff import *
    from numdiff import _reload_package

    def set_backend(BACKEND):
        numdiff.set_backend(BACKEND)
        _reload_package("klove")

    def get_backend():
        return numdiff.get_backend()

    BACKEND = numdiff.BACKEND

except:
    import numpy as backend

    def set_backend(BACKEND):
        pass

    def get_backend():
        return "numpy"

    BACKEND = "numpy"


from .bztools import *
from .core import *
from .viz import *



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove


def test_gif():
    import os

    import matplotlib.pyplot as plt

    from klove.viz import GifMaker

    plt.figure()

    gmk = GifMaker()

    for i in range(3):
        plt.plot(i, i**2, "or")
        gmk.snap()

    outfile = gmk.save()
    os.remove(outfile)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove


def test_bands():
    from math import pi

    import matplotlib.pyplot as plt

    import klove as kl

    bk = kl.backend
    a = 1
    q = a / (2 * 3**0.5)
    gamma = bk.array(pi / 6)
    positions = [
        (q * bk.cos(gamma), q * bk.sin(gamma)),
        (-q * bk.cos(gamma), -q * bk.sin(gamma)),
    ]
    lattice_vectors = (a, 0), (a * bk.cos(gamma * 2), a * bk.sin(gamma * 2))
    plate = kl.ElasticPlate(1, 0.1, 1, 0.0)
    k = (4 * pi) ** 2 * plate.omega0(a)
    m = 1
    masses = [m, m]
    stiff = [k, k]
    res_array = []
    for m, k, pos in zip(masses, stiff, positions):
        res_array.append(kl.Resonator(m, k, pos))

    bs = kl.BandsSimulation(plate, res_array, lattice_vectors)

    Npw = 4
    Nk = 5

    b0 = bk.pi / a
    Gamma_point = 0, 0
    K_point = 2 * b0 / 3, 2 * b0 / 3**0.5
    M_point = 0, 2 * b0 / 3**0.5
    sym_points = [Gamma_point, M_point, K_point]
    ks = kl.init_bands(sym_points, Nk)
    bands = []
    for K in ks:
        omegans = bs.eigensolve(K, Npw, return_modes=False, hermitian=False)
        bands.append(omegans)

    bands = bk.stack(bands).real
    plt.figure()
    kl.plot_bands(
        sym_points, Nk, bands, xtickslabels=[r"$\Gamma$", r"$M$", r"$K$", r"$\Gamma$"]
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove


def test_derivatives():
    import klove as kl

    def _gradient(f):
        df = bk.gradient(f)
        return df[0] if kl.get_backend() == "torch" else df

    bk = kl.backend

    res = kl.Resonator(1, 1, (0, 0))

    nf = 1000
    omega = bk.linspace(0.1, 0.9, nf)

    f = res.strength(omega)
    domega = _gradient(omega)

    fd = _gradient(f) / domega
    df = res.dstrength_domega(omega)

    nres = 5
    res_array = [kl.Resonator(1, 1, (i, 0)) for i in range(nres)]

    shift = 2
    tol = 1e-3
    assert bk.allclose(df[shift:-shift], fd[shift:-shift], rtol=tol)

    plate = kl.ElasticPlate(1, 2, 3, 0.1)

    for alternative in [True, False]:
        simu = kl.ScatteringSimulation(plate, res_array, alternative)

        k = simu.wavenumber(omega)
        fd = _gradient(k) / domega

        df = simu.dwavenumber_domega(omega)

        assert bk.allclose(df[shift:-shift], fd[shift:-shift], rtol=tol)

        s = simu._strength(omega, res)
        fd = _gradient(s) / domega

        df = simu._dstrength_domega(omega, res)
        assert bk.allclose(df[shift:-shift], fd[shift:-shift], rtol=tol)

        r = 0.1

        dk = _gradient(k)
        g = kl.special.greens_function(k, r)
        fd = _gradient(g) / dk

        df = kl.special.dgreens_function_dk(k, r)

        assert bk.allclose(df[shift:-shift], fd[shift:-shift], rtol=tol)

        M = simu.build_matrix(omega)
        dM = simu.build_matrix_derivative(omega)

        for i in range(simu.n_res):
            for j in range(simu.n_res):
                print(i, j)
                f = M[i, j]
                fd = _gradient(f) / domega
                df = dM[i, j]
                print(bk.mean(abs(df[shift:-shift] - fd[shift:-shift])))
                assert bk.allclose(df[shift:-shift], fd[shift:-shift], rtol=tol)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove

__all__ = ["init_bands", "init_bands_plot"]


from . import backend as bk


def init_bands(sym_points, nband):
    Gamma_point, X_point, M_point = sym_points
    _kx = bk.linspace(Gamma_point[0], X_point[0], nband)
    _ky = bk.linspace(Gamma_point[1], X_point[1], nband)
    kGammaX = bk.vstack([_kx, _ky])
    _kx = bk.linspace(X_point[0], M_point[0], nband)
    _ky = bk.linspace(X_point[1], M_point[1], nband)
    kXM = bk.vstack([_kx, _ky])
    _kx = bk.linspace(M_point[0], Gamma_point[0], nband)
    _ky = bk.linspace(M_point[1], Gamma_point[1], nband)
    kMGamma = bk.vstack([_kx, _ky])
    ks = bk.vstack([kGammaX[:, :-1].T, kXM[:, :-1].T, kMGamma.T])
    return ks


def init_bands_plot(sym_points, nband):
    Gamma_point, X_point, M_point = sym_points
    dMGamma = bk.linalg.norm(bk.array(Gamma_point) - bk.array(M_point))
    dGammaX = bk.linalg.norm(bk.array(X_point) - bk.array(Gamma_point))
    dXM = bk.linalg.norm(bk.array(M_point) - bk.array(X_point))
    _kx = bk.linspace(0, dGammaX, nband)[:-1]
    _kx1 = dGammaX + bk.linspace(0, dXM, nband)[:-1]
    _kx2 = dXM + dGammaX + bk.linspace(0, dMGamma, nband)
    ksplot = bk.hstack([_kx, _kx1, _kx2])
    kdist = [dGammaX, dXM, dMGamma]
    return ksplot, kdist



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove


def test_eig():
    import numpy as np

    from klove import backend as bk
    from klove import get_backend
    from klove.eig import polyeig

    np.random.seed(1234)
    N = 10
    P = 5
    A = bk.array(np.random.rand(P, N, N))
    e, X = polyeig(*A)
    # Test that eigenvector and value satisfy eigenvalue problem:
    for i in range(N):
        s = e[i]
        x = X[:, i]
        M = 0
        for p in range(P):
            M *= A[p] * s**p

        res = M @ x  # residuals
        assert bk.all(np.abs(res) < 1e-12)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove


__all__ = ["GifMaker", "plot_bands"]


import glob
import os
import tempfile
from itertools import accumulate

import matplotlib.pyplot as plt
from PIL import Image

from . import backend as bk
from .bztools import init_bands_plot


def save_frame(i, folder=".frames"):
    _name = f"{i}".zfill(3)
    filename = f"{folder}/{_name}.png"
    plt.savefig(filename)


def make_gif(folder=".frames", fps=10, name="animation"):
    imgs = sorted(glob.glob(f"{folder}/*.png"))
    img, *imgs = [Image.open(f, mode="r").convert("RGB") for f in imgs]
    out = img.save(
        f"{name}.gif",
        format="GIF",
        append_images=imgs,
        save_all=True,
        duration=1000 / fps,
        loop=0,
    )
    return os.path.abspath(f"{name}.gif")


class GifMaker:
    def __init__(self, folder=".frames"):
        self.folder = folder
        self.iteration = 0
        self.folder = tempfile.mkdtemp()

    def snap(self):
        self.iteration += 1
        save_frame(self.iteration, self.folder)

    def save(self, *args, **kwargs):
        return make_gif(self.folder, *args, **kwargs)


def plot_bands(
    sym_points,
    nband,
    eigenvalues,
    xtickslabels=[r"$\Gamma$", r"$X$", r"$M$", r"$\Gamma$"],
    color=None,
    **kwargs,
):
    # nband = int((len(eigenvalues)-2)/3)

    if color == None:
        color = "#4d63c5"
        if "color" in kwargs:
            kwargs.pop("colors")
        if "c" in kwargs:
            kwargs.pop("c")

    ksplot, kdist = init_bands_plot(sym_points, nband)
    plt.plot(ksplot, eigenvalues, color=color, **kwargs)
    # xticks = bk.cumsum(bk.array([0] + kdist))
    xticks = list(accumulate([0] + kdist))
    plt.xticks(xticks, xtickslabels)
    for x in xticks:
        plt.axvline(x, c="#8a8a8a")

    plt.xlim(xticks[0], xticks[-1])

    plt.ylabel(r"$\omega$")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove


import importlib.metadata as metadata


def get_meta(metadata):
    data = metadata.metadata("klove")
    __version__ = metadata.version("klove")
    __author__ = data.get("author")
    __description__ = data.get("summary")
    return __version__, __author__, __description__, data


__version__, __author__, __description__, data = get_meta(metadata)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove


def test_diff():
    import matplotlib.pyplot as plt

    import klove as kl

    plt.ion()
    plt.close("all")

    bk = kl.backend

    or0 = 0.91
    m0 = 3e-4
    k0 = or0**2 * m0
    Nres = 10
    period = 1.0
    t = bk.linspace(0, 2 * bk.pi, Nres + 1)[:-1]
    R0 = period / 2 * 0.8
    xpos = period / 2 + R0 * bk.cos(t)
    ypos = R0 * bk.sin(t)
    resarray = []
    for xp, yp in zip(xpos, ypos):
        res = kl.Resonator(m0, k0, (xp, yp))
        resarray.append(res)
    plate = kl.ElasticPlate(0.1, 10, 1, 0.1)

    nh = 0

    simu = kl.DiffractionSimulation(plate, resarray, period, nh)

    omega = 0.3
    angle = bk.pi / 2  # - bk.pi / 6

    simu.nh = simu.get_min_harmonics(omega, angle)

    sol = simu.solve(omega, angle)
    print(sol)

    ny = 4
    npx = 100
    npy = npx * ny
    x1 = bk.linspace(0, period, npx)
    y1 = bk.linspace(-ny * period, ny * period, npy)
    x, y = bk.meshgrid(x1, y1, indexing="xy")
    W = simu.get_field(x, y, sol, omega, angle)
    Ws = simu.get_scattered_field(x, y, sol, omega, angle)

    # fig,ax=plt.subplots(1)
    # field = bk.abs(W)
    # plt.pcolormesh(x, y,field , cmap="inferno")
    # plt.gca().set_aspect("equal")
    # for res in resarray:
    #     plt.plot(*res.position, ".w")
    # plt.colorbar()
    # plt.suptitle("total")

    # fig,ax=plt.subplots(1)
    # field = bk.abs(Ws)
    # plt.pcolormesh(x, y,field , cmap="inferno")
    # plt.gca().set_aspect("equal")
    # for res in resarray:
    #     plt.plot(*res.position, ".w")
    # plt.colorbar()
    # plt.suptitle("scattered")

    # fig,ax=plt.subplots(1,2)
    # for i in range(2):
    #     plt.sca(ax[i])
    #     field = bk.real(W) if i==0 else bk.imag(W)
    #     plt.pcolormesh(x, y,field , cmap="RdBu_r")
    #     plt.gca().set_aspect("equal")
    #     for res in resarray:
    #         plt.plot(*res.position, ".k")
    #     plt.colorbar()

    # plt.suptitle("total")

    # fig,ax=plt.subplots(1,2)
    # for i in range(2):
    #     plt.sca(ax[i])
    #     field = bk.real(Ws) if i==0 else bk.imag(Ws)
    #     plt.pcolormesh(x, y, field, cmap="RdBu_r")
    #     plt.gca().set_aspect("equal")
    #     for res in resarray:
    #         plt.plot(*res.position, ".k")
    #     plt.colorbar()

    # plt.suptitle("scattered")

    R, T = simu.get_efficiencies(sol, omega, angle)
    print("Reflection")
    print("----------")
    print(R["energy"])
    print(R["total"])
    print("Transmission")
    print("----------")
    print(T["energy"])
    print(T["total"])
    print("Balance")
    print("----------")
    balance = R["total"] + T["total"]
    print(balance)
    assert abs(balance - 1) < 1e-4

    omega_max = 0.4

    nh_ = simu.get_min_harmonics(omega_max, angle)
    print("max nh", nh_)
    simu.nh = nh
    simu.nh = 1
    # sys.exit(0)

    omegas = bk.linspace(0.01, omega_max, 100)
    Rs = []
    Ts = []
    for omega in omegas:
        sol = simu.solve(omega, angle)
        R, T = simu.get_efficiencies(sol, omega, angle)
        Rs.append(bk.array(list(R["energy"].values())))
        Ts.append(bk.array(list(T["energy"].values())))

    # plt.figure()
    # plt.plot(omegas, Rs)
    # plt.plot(omegas, Ts, "--")
    Rtot = bk.array([sum(_) for _ in Rs])
    Ttot = bk.array([sum(_) for _ in Ts])
    Balance = Rtot + Ttot
    # plt.plot(omegas, Rtot, "b")
    # plt.plot(omegas, Ttot, "--r")
    # plt.plot(omegas, Balance, "-k")

    for b in Balance:
        assert abs(b - 1) < 1e-4



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove


def test_alt():
    import klove as kl

    res0 = kl.Resonator(1, 1, (0, 0))
    res1 = kl.Resonator(3, 2, (-1, -3))
    plate = kl.ElasticPlate(0.7, 22, 3, 0.1)
    resarray = [res0, res1]
    simu0 = kl.ScatteringSimulation(plate, resarray, alternative=False)

    omega = 2
    pos_src = (1, 1)

    sol0 = simu0.solve(simu0.point_source, omega, pos_src)

    simu1 = kl.ScatteringSimulation(plate, resarray, alternative=True)
    sol1 = simu1.solve(simu1.point_source, omega, pos_src)

    pos_probe = (0.2, 0.4)

    w0 = simu0.get_scattered_field(*pos_probe, sol0, omega)

    w1 = simu1.get_scattered_field(*pos_probe, sol1, omega)

    assert abs(w0 - w1) < 1e-6



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove


import scipy.special as sp

from . import backend as bk
from . import get_backend

BACKEND = get_backend()

EPS = 1e-12


def j0(kr):
    return sp.jn(0, kr)


def j1(kr):
    return sp.jn(1, kr)


def y0(kr):
    return sp.yv(0, kr)


def y1(kr):
    return sp.yv(1, kr)


def k0(kr):
    return sp.kv(0, kr)


def k1(kr):
    return sp.kv(1, kr)


def j2(kr):
    return sp.jn(2, kr)


def y2(kr):
    return sp.yv(2, kr)


def k2(kr):
    return sp.kv(2, kr)


def j3(kr):
    return sp.jn(3, kr)


def y3(kr):
    return sp.yv(3, kr)


def k3(kr):
    return sp.kv(3, kr)


def torch_j2(kr):
    return 2 / kr * bk.special.bessel_j1(kr) - bk.special.bessel_j0(kr)


def torch_y2(kr):
    return 2 / kr * bk.special.bessel_y1(kr) - bk.special.bessel_y0(kr)


def torch_k2(kr):
    return 2 / kr * bk.special.modified_bessel_k1(kr) + bk.special.modified_bessel_k0(
        kr
    )


def torch_j3(kr):
    return 4 / kr * torch_j2(kr) - bk.special.bessel_j1(kr)


def torch_y3(kr):
    return 4 / kr * torch_y2(kr) - bk.special.bessel_y1(kr)


def torch_k3(kr):
    return 4 / kr * torch_k2(kr) + bk.special.modified_bessel_k1(kr)


bessel_j0 = bk.special.bessel_j0 if BACKEND == "torch" else sp.j0
bessel_y0 = bk.special.bessel_y0 if BACKEND == "torch" else sp.y0
bessel_k0 = bk.special.modified_bessel_k0 if BACKEND == "torch" else sp.k0
bessel_j1 = bk.special.bessel_j1 if BACKEND == "torch" else sp.j1
bessel_y1 = bk.special.bessel_y1 if BACKEND == "torch" else sp.y1
bessel_k1 = bk.special.modified_bessel_k1 if BACKEND == "torch" else sp.k1
bessel_j2 = torch_j2 if BACKEND == "torch" else j2
bessel_y2 = torch_y2 if BACKEND == "torch" else y2
bessel_k2 = torch_k2 if BACKEND == "torch" else k2
bessel_j3 = torch_j3 if BACKEND == "torch" else j3
bessel_y3 = torch_y3 if BACKEND == "torch" else y3
bessel_k3 = torch_k3 if BACKEND == "torch" else k3


def switch_complex(f1, f2):
    def wrapper(z):
        z = bk.array(z)
        cond = bk.is_complex(z) if BACKEND == "torch" else bk.any(bk.iscomplex(z))
        if cond:
            return f2(z)
        return f1(z.real)

    return wrapper


bessel_j0 = switch_complex(bessel_j0, j0)
bessel_y0 = switch_complex(bessel_y0, y0)
bessel_k0 = switch_complex(bessel_k0, k0)
bessel_j1 = switch_complex(bessel_j1, j1)
bessel_y1 = switch_complex(bessel_y1, y1)
bessel_k1 = switch_complex(bessel_k1, k1)
bessel_j2 = switch_complex(bessel_j2, j2)
bessel_y2 = switch_complex(bessel_y2, y2)
bessel_k2 = switch_complex(bessel_k2, k2)
bessel_j3 = switch_complex(bessel_j3, j3)
bessel_y3 = switch_complex(bessel_y3, y3)
bessel_k3 = switch_complex(bessel_k3, k3)


def _gfreal(kr):
    return bk.where(bk.abs(bk.array(kr)) < EPS, 1.0, bessel_j0(kr))


def _gfimag(kr):
    return bk.where(
        bk.abs(bk.array(kr)) < EPS,
        0.0,
        bessel_y0(kr) + 2 / bk.pi * bessel_k0(kr),
    )


def _dgfreal(kr):
    return bk.where(bk.abs(bk.array(kr)) < EPS, 0.0, -bessel_j1(kr))


def _dgfimag(kr):
    return bk.where(
        bk.abs(bk.array(kr)) < EPS,
        0.0,
        -bessel_y1(kr) - 2 / bk.pi * bessel_k1(kr),
    )


if BACKEND == "torch":

    class _GFreal(bk.autograd.Function):
        @staticmethod
        def forward(ctx, kr):
            ctx.save_for_backward(kr)
            return _gfreal(kr)

        @staticmethod
        def backward(ctx, grad_output):
            (kr,) = ctx.saved_tensors
            return grad_output * _dgfreal(kr)

    class _GFimag(bk.autograd.Function):
        @staticmethod
        def forward(ctx, kr):
            ctx.save_for_backward(kr)
            return _gfimag(kr)

        @staticmethod
        def backward(ctx, grad_output):
            (kr,) = ctx.saved_tensors
            return grad_output * _dgfimag(kr)

    gfreal = _GFreal.apply
    gfimag = _GFimag.apply


else:
    gfreal = _gfreal
    gfimag = _gfimag


def _norma_gf(k):
    return 1j / (8 * k**2)


def greens_function(k, r):
    kr = k * r
    norm = _norma_gf(k)
    return norm * (gfreal(kr) + 1j * gfimag(kr))


def dgreens_function_dk(k, r):
    G0 = greens_function(k, r)
    G1 = _gf_fun(k, r, bessel_j1, bessel_y1, bessel_k1)
    return -2 / k * G0 - G1 * r


def greens_function_cartesian(k, x, y):
    r = radial_coordinates(x, y)
    return greens_function(k, r)


def radial_coordinates(x, y):
    return (EPS + x**2 + y**2) ** 0.5


def _gf_fun(k, r, jfun, yfun, kfun):
    kr = k * r
    re = jfun(kr)
    im = yfun(kr) + 2 / bk.pi * kfun(kr)
    norm = _norma_gf(k)
    return norm * (re + 1j * im)


def grad_greens_function_cartesian(k, x, y):
    r = radial_coordinates(x, y)
    out = -k / r * _gf_fun(k, r, bessel_j1, bessel_y1, bessel_k1)
    out = bk.where(bk.abs(k * r) < EPS, 0.0 + 0.0 * 1j, out)
    return x * out, y * out


def laplacian_greens_function_cartesian(k, x, y):
    r = radial_coordinates(x, y)
    G1 = _gf_fun(k, r, bessel_j1, bessel_y1, bessel_k1)
    G2 = _gf_fun(k, r, bessel_j2, bessel_y2, bessel_k2)
    out = -2 * k / r * G1 + k**2 * G2
    out = bk.where(bk.abs(k * r) < EPS, 0.0 + 0.0 * 1j, out)
    return out


def grad_laplacian_greens_function_cartesian(k, x, y):
    r = radial_coordinates(x, y)
    G2 = _gf_fun(k, r, bessel_j2, bessel_y2, bessel_k2)
    G3 = _gf_fun(k, r, bessel_j3, bessel_y3, bessel_k3)

    def tmp(x):
        out = 4 * k**2 * x / r**2 * G2 - k**3 * x / r * G3
        out = bk.where(bk.abs(k * r) < EPS, 0.0 + 0.0 * 1j, out)
        return out

    return tmp(x), tmp(y)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove

__all__ = ["nonlinear_eigensolver"]

import logging

import matplotlib.pyplot as plt
import scipy.linalg as la
from skimage.feature import peak_local_max

from . import backend as bk
from . import get_backend


def get_residual(M, phi):
    return bk.abs(phi @ M @ phi)


def eig(C, D):
    if get_backend() == "torch":
        invD = bk.linalg.inv(D)
        return bk.linalg.eig(invD @ C)
    return la.eig(C, D)


class ConvergenceError(Exception):
    pass


class EigenvalueError(Exception):
    pass


def null(M, mult=1):
    e, v = bk.linalg.eig(M)
    srt = bk.argsort(bk.abs(e))
    return v[:, srt[0:mult]]


def block(matrices):
    if get_backend() == "torch":
        return bk.cat([bk.cat(m, dim=1) for m in matrices], dim=0)
    return bk.block(matrices)


def _dot(a, b):
    if get_backend() == "torch":
        return bk.matmul(a, b)
    else:
        return bk.dot(a, b)


def eig_newton(
    func,
    lambda_0,
    x_0,
    lambda_tol=1e-8,
    max_iter=20,
    func_gives_der=False,
    G=None,
    args=[],
    weight="rayleigh symmetric",
    y_0=None,
):
    """Solve a nonlinear eigenvalue problem by Newton iteration

    Parameters
    ----------
    func : function
        The function with input `lambda` which returns the matrix
    lambda_0 : complex
        The starting guess for the eigenvalue
    x_0 : ndarray
        The starting guess for the eigenvector
    lambda_tol : float
        The relative tolerance in the eigenvalue for convergence
    max_iter : int
        The maximum number of iterations to perform
    func_gives_der : boolean, optional
        If `True`, then the function also returns the derivative as the second
        returned value. If `False` finite differences will be used instead,
        which will have reduced accuracy
    args : list, optional
        Any additional arguments to be supplied to `func`
    weight : string, optional
        How to perform the weighting of the eigenvector

        'max element' : The element with largest magnitude will be preserved

        'rayleigh' : Rayleigh iteration for Hermitian matrices will be used

        'rayleigh symmetric' : Rayleigh iteration for complex symmetric
        (i.e. non-Hermitian) matrices will be used

        'rayleigh asymmetric' : Rayleigh iteration for general matrices

    y_0 : ndarray, optional
        For 'rayleigh asymmetric weighting', this is required as the initial
        guess for the left eigenvector

    Returns
    -------
    res : dictionary
        A dictionary containing the following members:

        `eigval` : the eigenvalue

        'eigvec' : the eigenvector

        'iter_count' : the number of iterations performed

        'delta_lambda' : the change in the eigenvalue on the final iteration


    See:
    1.  P. Lancaster, Lambda Matrices and Vibrating Systems.
        Oxford: Pergamon, 1966.

    2.  A. Ruhe, “Algorithms for the Nonlinear Eigenvalue Problem,”
        SIAM J. Numer. Anal., vol. 10, no. 4, pp. 674–689, Sep. 1973.

    """

    x_s = x_0
    lambda_s = lambda_0

    if weight.lower() == "rayleigh asymmetric":
        if y_0 is None:
            raise ValueError("Parameter y_0 must be supplied for asymmetric " "case")
        y_s = y_0

    logging.debug("Searching for zeros with eig_newton")
    logging.debug("Starting guess %+.4e %+.4ej" % (lambda_0.real, lambda_0.imag))

    converged = False

    if not func_gives_der:
        # evaluate at an arbitrary nearby starting point to allow finite
        # differences to be taken
        # lambda_sm = lambda_0*(1+10j*lambda_tol)
        step = bk.max(1e-6, lambda_tol * 10)
        step = lambda_tol * 10
        lambda_sm = lambda_0 * (1 + (1 + 1j) * step)
        T_sm = func(lambda_sm, *args)

    for iter_count in range(max_iter):
        if func_gives_der:
            T_s, T_ds = func(lambda_s, *args)
        else:
            T_s = func(lambda_s, *args)
            T_ds = (T_s - T_sm) / (lambda_s - lambda_sm)

        # print(T_ds)
        w = _dot(T_ds, x_s)
        if get_backend() == "torch":
            T_s_lu = bk.linalg.lu_factor(T_s)

            u = bk.linalg.lu_solve(*T_s_lu, bk.stack([w]).T).flatten()

        else:
            T_s_lu = la.lu_factor(T_s)

            u = la.lu_solve(T_s_lu, _dot(T_ds, x_s))

        # if known_vects is supplied, we should take this into account when
        # finding v
        if weight.lower() == "max element":
            v_s = bk.zeros_like(x_s)
            v_s[bk.argmax(bk.abs(x_s))] = 1.0
        # elif weight.lower() == "rayleigh":
        #     v_s = _dot(T_s.T, x_s.conj())
        # elif weight.lower() == "rayleigh symmetric":
        #     v_s = _dot(T_s.T, x_s)
        # elif weight.lower() == "rayleigh asymmetric":
        #     y_s = bk.linalg.lu_solve(T_s_lu, _dot(T_ds.T, y_s), trans=1)

        #     y_s /= bk.sqrt(bk.sum(bk.abs(y_s) ** 2))
        #     v_s = _dot(T_s.T, y_s)
        else:
            raise ValueError("Unknown weighting method %s" % weight)

        delta_lambda_abs = _dot(v_s, x_s) / _dot(v_s, u)

        delta_lambda = bk.abs(delta_lambda_abs / lambda_s)
        converged = delta_lambda < lambda_tol

        if converged:
            break

        lambda_s1 = lambda_s - delta_lambda_abs
        x_s1 = u / bk.sqrt(bk.sum(bk.abs(u) ** 2))

        # update variables for next iteration
        if not func_gives_der:
            lambda_sm = lambda_s
            T_sm = T_s

        lambda_s = lambda_s1
        x_s = x_s1
        logging.debug("%+.4e %+.4ej" % (lambda_s.real, lambda_s.imag))

    if not converged:
        return
        # raise ConvergenceError("maximum iterations reached, no convergence")

    res = {
        "eigval": lambda_s,
        "iter_count": iter_count + 1,
        "delta_lambda": delta_lambda,
    }

    if weight.lower() == "rayleigh asymmetric":
        # Scale both the left and right eigenvector identically first
        y_s /= bk.sqrt(bk.vdot(y_s, y_s) / bk.vdot(x_s, x_s))

        # Then scale both to match the eigenvalue derivative
        dz_ds = _dot(y_s, _dot(T_ds, x_s))
        y_s /= bk.sqrt(dz_ds)
        res["eigvec_left"] = y_s

        x_s /= bk.sqrt(dz_ds)
        res["eigvec"] = x_s

    else:
        # scale the eigenvector so that the eigenvalue derivative is 1
        dz_ds = _dot(x_s, _dot(T_ds, x_s))
        x_s /= bk.sqrt(dz_ds)
        res["eigvec"] = x_s
        res["eigvec_left"] = x_s

    return res


def polyeig(*A):
    """Solve the polynomial eigenvalue problem: (A0 + e A1 +...+  e**p Ap)x=0

    Parameters
    ----------
    A : list of arrays
        The arrays for each power of the polynomial (in increasing order)

    Returns
    -------
    tuple of arrays
        eigenvalues e and eigenvectors x

    """
    if len(A) <= 0:
        raise Exception("Provide at least one matrix")
    for Ai in A:
        if Ai.shape[0] != Ai.shape[1]:
            raise Exception("Matrices must be square")
        if Ai.shape != A[0].shape:
            raise Exception("All matrices must have the same shapes")

    n = A[0].shape[0]
    l = len(A) - 1
    # Assemble matrices for generalized problem
    C = block(
        [[bk.zeros((n * (l - 1), n)), bk.eye(n * (l - 1))], [-bk.column_stack(A[0:-1])]]
    )
    D = block(
        [
            [bk.eye(n * (l - 1)), bk.zeros((n * (l - 1), n))],
            [bk.zeros((n, n * (l - 1))), A[-1]],
        ]
    )
    # Solve generalized eigenvalue problem
    e, X = eig(C, D)
    if bk.all(bk.isreal(e)):
        e = bk.real(e)
    X = X[:n, :]

    # Sort eigenvalues/vectors
    # I = bk.argsort(e)
    # X = X[:,I]
    # e = e[I]

    # Scaling each mode by max
    maxi = bk.max(bk.abs(X), axis=0)

    if get_backend() == "torch":
        maxi = maxi[0]
    X /= bk.tile(maxi, (n, 1))

    return e, X


def nonlinear_eigensolver(
    simu,
    omega0,
    omega1,
    guesses=None,
    weight="max element",
    init_vect="eig",
    strategy="peaks",
    func_gives_der="True",
    lambda_tol=1e-6,
    max_iter=100,
    N_grid=(10, 10),
    N_guess_loc=1,
    Rloc=0.1,
    plot_solver=False,
    peak_ref=10,
    verbose=False,
):
    Nguess_re, Nguess_im = N_grid

    if func_gives_der:

        def func_eig(omega):
            return simu.build_matrix(omega), simu.build_matrix_derivative(omega)

    else:

        def func_eig(omega):
            return simu.build_matrix(omega)

    if guesses is None:
        if strategy == "grid":
            guesses_re = bk.linspace(omega0.real, omega1.real, Nguess_re)
            guesses_im = bk.linspace(omega0.imag, omega1.imag, Nguess_im)
            guesses_re, guesses_im = bk.meshgrid(guesses_re, guesses_im)
            guesses = guesses_re + 1j * guesses_im
            guesses = guesses.flatten()
        else:
            Nre, Nim = Nguess_re * peak_ref, Nguess_im * peak_ref
            omegas_re = bk.linspace(omega0.real, omega1.real, Nre)
            omegas_im = bk.linspace(omega1.imag, omega0.imag, Nim)
            omegas_re_, omegas_im_ = bk.meshgrid(omegas_re, omegas_im, indexing="ij")

            #################################################################
            # Compute complex plane quantities

            omegas_complex = omegas_re_ + 1j * omegas_im_
            Mc = simu.build_matrix(omegas_complex)
            if get_backend() == "torch":
                Mc = bk.permute(Mc, (2, 3, 0, 1))
            else:
                Mc = bk.transpose(Mc, axes=(2, 3, 0, 1))
            evs = bk.linalg.eigvals(Mc)
            srt = bk.argsort(bk.abs(evs), axis=-1)
            if get_backend() == "torch":
                min_evs = bk.gather(evs, -1, srt)[:, :, 0]
            else:
                min_evs = bk.take_along_axis(evs, srt, axis=-1)[:, :, 0]

            im = -bk.log10(bk.abs(min_evs))
            if get_backend() == "torch":
                im = im.numpy()

            coordinates = peak_local_max(im, min_distance=1)

            guess_peak = bk.array([omegas_complex[*coord] for coord in coordinates])

            tloc = bk.linspace(0, 2 * bk.pi, N_guess_loc + 1)[:-1]
            guesses = []
            for guess_loc in guess_peak:
                guesses_ = (
                    guess_loc.real
                    + Rloc * bk.cos(tloc)
                    + 1j * (guess_loc.imag + Rloc * bk.sin(tloc))
                )
                guesses_ = bk.hstack([guesses_, guess_loc])
                guesses.append(guesses_)
            guesses = bk.stack(guesses).flatten()

    if plot_solver:
        plt.plot(guesses.real, guesses.imag, ".k")
        plt.pause(0.001)

    def compute_eigenvalues(guesses, lambda_tol, max_iter, filter=True):
        evs = []
        modes = []
        residuals = []
        for guess in guesses:
            if init_vect == "eig":
                vect_init = null(simu.build_matrix(guess))[:, 0]
            else:
                vect_init = bk.random.rand(simu.n_res)

            res = eig_newton(
                func_eig,
                guess,
                vect_init,
                lambda_tol=lambda_tol,
                max_iter=max_iter,
                func_gives_der=func_gives_der,
                weight=weight,
            )
            if res is not None:
                ev = res["eigval"]
                # print(res["iter_count"],res["delta_lambda"])
                if (
                    ev.real > omega0.real
                    and ev.real < omega1.real
                    and ev.imag < omega1.imag
                    and ev.imag > omega0.imag
                ):
                    evs.append(ev)
                    modes.append(res["eigvec"])
                    residuals.append(res["delta_lambda"])
                    if plot_solver:
                        plt.plot(ev.real, ev.imag, ".r")
                        plt.xlim(omega0.real, omega1.real)
                        plt.ylim(omega0.imag, omega1.imag)
                        plt.pause(0.001)

        if evs == []:
            raise EigenvalueError("No eigenvalues found")
        evs = bk.array(evs)
        modes = bk.stack(modes).T
        residuals = bk.array(residuals)
        if filter:
            precision = lambda_tol * 100
            res_unique = []
            evs_unique = []
            modes_unique = []
            for i, ev in enumerate(evs):
                evfloor = (
                    bk.floor(ev.real / precision) + 1j * bk.floor(ev.imag / precision)
                ) * precision
                if not (evfloor in evs_unique):
                    evs_unique.append(evfloor)
                    res_unique.append(residuals[i])
                    modes_unique.append(modes[:, i])

            modes = bk.stack(modes_unique).T
            evs = bk.array(evs_unique)
            residuals = bk.array(res_unique)
        if verbose:
            print_results(evs, residuals)
        return evs, modes, residuals

    evs, modes, residuals = compute_eigenvalues(
        guesses, lambda_tol, max_iter, filter=True
    )
    return evs, modes


def print_results(evs0, res0, message=None):
    print("")
    print("----------------------------------------------------")
    if message is not None:
        print(message)
    print(f"Found {len(evs0)} eigenvalues")
    print("----------------------------------------------------")
    print("              eigenvalue                 residual")
    print("----------------------------------------------------")
    for z0, r0 in zip(evs0, res0):
        space = " " if z0.real > 0 else ""
        print(f"{space}{z0:.14f}     {r0:.3e}")


# def gram_schmidt(A):
#     """Orthogonalize a set of vectors stored as the columns of matrix A."""
#     # Get the number of vectors.
#     n = A.shape[1]
#     for j in range(n):
#         # To orthogonalize the vector in column j with respect to the
#         # previous vectors, subtract from it its projection onto
#         # each of the previous vectors.
#         for k in range(j):
#             A[:, j] -= bk.dot(A[:, k], A[:, j]) * A[:, k]
#         A[:, j] = A[:, j] / bk.linalg.norm(A[:, j])
#     return A


# def _from_basis_to_project_matrix(B):
#     """Formula page 87 book 'Maths For ML'"""
#     inv_Bt_B = bk.linalg.inv(bk.dot(B.T, B))
#     proj_matrix = bk.dot(B, bk.dot(inv_Bt_B, B.T))
#     return proj_matrix


# def _projec(x, basis):
#     """project `x` in the new vector space defined by `basis`"""
#     proj_matrix = _from_basis_to_project_matrix(basis)
#     proj_x = bk.dot(proj_matrix, x)
#     return proj_x


# def _get_col(x, col_id):
#     """return column `col_id` from matrix `x`"""
#     raw_col_val = x[:, col_id]
#     col_as_row = bk.array([raw_col_val]).T
#     return col_as_row


# def gram_schmidt(B):
#     nb_col = B.shape[1]
#     first_col = _get_col(B, 0)

#     U_vecs = [first_col]

#     for i in range(1, nb_col):
#         B_i = _get_col(B, i)
#         U_i_1 = bk.concatenate(U_vecs, axis=1)
#         p = _projec(B_i, U_i_1)
#         U_i = B_i - p
#         U_vecs.append(U_i)

#     return U_vecs



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove


def test_far_field():
    from math import pi

    import klove as kl

    bk = kl.backend

    plate = kl.ElasticPlate(h=1, rho=1, E=1, nu=0.21)

    omega0 = 2e3
    a = 1
    R = 1
    Nres = 6

    m0 = plate.rho * a**2 * plate.h
    ms = bk.ones(Nres) * m0
    alpha = 2
    dist = bk.linspace(0.5, 1.5, Nres)
    ks = m0 * omega0**2 * dist**2
    t0 = -69.81 * bk.ones(Nres) * plate.bending_stiffness

    resarray = []
    for alpha in range(Nres):
        q = bk.array(2 * pi * alpha / Nres)
        xalpha = R * bk.cos(q)
        yalpha = R * bk.sin(q)
        resarray.append(kl.ConstantStrength(t0[alpha], (xalpha, yalpha)))

    sim = kl.ScatteringSimulation(plate, resarray, alternative=True)

    Nomegas = 5
    k = bk.linspace(0.001, 8, Nomegas)
    omegas = k**2 / (plate.rho * plate.h / plate.bending_stiffness) ** 0.5
    Ntheta = 3
    angles = bk.linspace(0, 2 * pi, Ntheta)
    far_field = bk.zeros([Nomegas, Ntheta], dtype=bk.complex64)
    sigma_sc = bk.zeros(Nomegas)
    for omegaIterator, omega in enumerate(omegas):
        incident = sim.plane_wave
        phi_n = sim.solve(incident, omega, angle=0)
        far_field[omegaIterator, :] = sim.get_far_field_radiation(phi_n, omega, angles)
        sigma_sc[omegaIterator] = sim.get_scattering_cross_section(phi_n, omega)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove


def test_qnms():
    from math import pi

    import klove as kl

    bk = kl.backend
    Nres = 3
    alternative = True
    a = 1
    plate = kl.ElasticPlate(a / 10, 1, 1, 0.3)
    omega_plate0 = plate.omega0(a)
    m_plate = plate.rho * a**2 * plate.h
    omegar0 = 1
    m0 = m_plate / 10
    k0 = omegar0**2 * m0
    positionsx = bk.linspace(-Nres * a / 2, Nres * a / 2, Nres)
    positions = [(_x, 0) for _x in positionsx]
    masses = bk.ones(Nres) * m0
    stiffnesses = bk.ones(Nres) * k0
    res_array = []
    for i in range(Nres):
        res_array.append(kl.Resonator(masses[i], stiffnesses[i], positions[i]))
    simu = kl.ScatteringSimulation(plate, res_array, alternative=alternative)

    singularity_Talpha = simu.get_strength_poles()
    omega0 = 0.1 - 1.13 * 1j
    omega1 = 1.91 - 1e-14 * 1j

    N_guess_loc = 3
    lambda_tol = 1e-12
    max_iter = 100
    evs, eigenvectors = simu.eigensolve(
        omega0,
        omega1,
        lambda_tol=lambda_tol,
        max_iter=max_iter,
        N_guess_loc=N_guess_loc,
        Rloc=0.01,
    )

    Nmodes = len(evs)

    M = [simu.build_matrix(omega) for omega in evs]
    dM = [simu.build_matrix_derivative(omega) for omega in evs]

    multiplicities = simu.get_multiplicities(evs, M)

    evs, eigenvectors = simu.update_eig(evs, eigenvectors, multiplicities)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove


def test_env_var():
    import klove as kl

    backend = kl.get_backend()
    print(backend)


def run(kl, backend):
    res0 = kl.Resonator(1, 1, (0, 0))
    res1 = kl.Resonator(3, 2, (-1, -3))
    plate = kl.ElasticPlate(0.7, 22, 3, 0.1)
    resarray = [res0, res1]
    simu0 = kl.ScatteringSimulation(plate, resarray, alternative=False)
    omega = 2
    pos_src = (1, 1)
    sol0 = simu0.solve(simu0.point_source, omega, pos_src)
    pos_probe = (0.2, 0.4)
    w0 = simu0.get_scattered_field(*pos_probe, sol0, omega)
    return sol0, w0


# def test_backends():
#     import numpy as npo

#     import klove as kl

#     for backend in kl.numdiff.available_backends:
#         print(backend)
#         kl.set_backend(backend)

#         bk = kl.backend
#         assert kl.get_backend() == backend

#         if backend != "jax":
#             sol, w = run(kl, backend)

#             if backend == "numpy":
#                 sol0 = sol.copy()
#                 w0 = w.copy()
#             else:
#                 assert bk.allclose(bk.array(w0), w)
#                 assert bk.allclose(bk.array(sol0), sol)
#             # TODO: Implement jax version

#     kl.set_backend("numpy")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove


def test_import_nonumdiff():
    import sys

    import klove

    sys.modules["numdiff"] = None
    import importlib

    importlib.reload(klove)

    assert klove.set_backend("Whatever") is None
    assert klove.get_backend() == "numpy"

    del sys.modules["numdiff"]
    import numdiff

    importlib.reload(klove)

    # klove.set_backend("numpy")



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Author: Benjamin Vial
# This file is part of klove
# License: GPLv3
# See the documentation at benvial.gitlab.io/klove


"""Core functions and classes.
"""


__all__ = [
    "Pin",
    "Mass",
    "Resonator",
    "ConstantStrength",
    "Beam",
    "ElasticPlate",
    "DiffractionSimulation",
    "ScatteringSimulation",
    "BandsSimulation",
]


from abc import ABC, abstractmethod
from dataclasses import dataclass

from . import backend as bk
from .eig import nonlinear_eigensolver
from .special import (
    dgreens_function_dk,
    greens_function,
    greens_function_cartesian,
    radial_coordinates,
)

_BIG = 1e12


class Scatterer(ABC):
    """Base class representing a scatterer."""

    @abstractmethod
    def strength(self):
        pass

    @abstractmethod
    def dstrength_domega(self):
        pass


@dataclass
class Pin(Scatterer):
    """Class representing a pin.

    Parameters
    ----------
    position : tuple
        The position of the pin on the plate.

    """

    position: tuple  #: The position

    def strength(self, omega) -> float:
        """Strength function of frequency `omega`"""
        # t_alpha * D in [Torrent2013]
        return _BIG

    def dstrength_domega(self, omega) -> float:
        """Derivative of strength function w.r.t. `omega`"""
        return 0


@dataclass
class Mass(Scatterer):
    """Class representing a mass.

    Parameters
    ----------
    mass : float
        Value of the mass.
    position : tuple
        The position of the mass on the plate.

    """

    mass: float
    position: tuple

    def strength(self, omega) -> float:
        """Strength function of frequency `omega`"""
        # t_alpha * D in [Torrent2013]
        return self.mass * omega**2

    def dstrength_domega(self, omega) -> float:
        """Derivative of strength function w.r.t. `omega`"""
        return self.mass * omega * 2


@dataclass
class Resonator(Scatterer):
    """Class representing a resonator.

    Parameters
    ----------
    mass : float
        Value of the mass.
    stiffness : float
        Stiffness constant of the spring.
    position : tuple
        The position of the resonator on the plate.

    """

    mass: float
    stiffness: float
    position: tuple

    @property
    def omega_r(self) -> float:
        """Resonant frequency."""
        return (self.stiffness / self.mass) ** 0.5

    def strength(self, omega) -> float:
        """Strength function of frequency `omega`"""
        # t_alpha * D in [Torrent2013]
        return (
            self.mass
            * self.omega_r**2
            * omega**2
            / (self.omega_r**2 - omega**2)
        )

    def dstrength_domega(self, omega) -> float:
        """Derivative of strength function w.r.t. `omega`"""
        return (
            2
            * self.mass
            * self.omega_r**4
            * omega
            / (self.omega_r**2 - omega**2) ** 2
        )


@dataclass
class ConstantStrength:
    """Class representing a scatterer with constant strength.

    Parameters
    ----------
    strength : float
        Value of the strength.
    position : tuple
        The position of the resonator on the plate.

    """

    impedance: float
    position: tuple

    # @property
    def strength(self, omega=0) -> float:
        # t_alpha * D in [Torrent2013]
        return self.impedance

    def dstrength_domega(self, omega=0) -> float:
        """Derivative of strength function w.r.t. `omega`"""
        return 0


@dataclass
class Beam:
    """Class representing a beam.

    Parameters
    ----------
    length : float
        Length of the beam.
    area : float
        Area of the cross section.
    rho : float
        Mass density.
    E : float
        Young's modulus
    position : tuple
        The position of the beam on the plate.

    """

    length: float
    area: float
    rho: float
    E: float
    position: tuple

    def strength(self, omega) -> float:
        """Strength function of frequency `omega`"""
        c = (self.E / self.rho) ** 0.5
        k = omega / c

        return self.rho * self.area * k * c**2 * bk.tan(bk.array(k * self.length))

    def dstrength_domega(self, omega) -> float:
        """Strength function of frequency `omega`"""
        c = (self.E / self.rho) ** 0.5
        k = omega / c
        kl = bk.array(k * self.length)

        return (
            self.rho
            * self.area
            * c**2
            * (bk.tan(kl) + k * self.length / cos(kl) ** 2)
        )


@dataclass
class ElasticPlate:
    """Class representing a thin elastic plate.

    Parameters
    ----------
    h : float
        Thickness
    rho : float
        Mass density
    E : float
        Young's modulus
    nu: float
        Poisson ratio
    """

    h: float  #: thickness
    rho: float  #: mass density
    E: float  #: Young's modulus
    nu: float  #: Poisson ratio

    @property
    def bending_stiffness(self) -> float:
        r"""The bending stiffness of the plate defined
        as :math:`D = E h^3 /(12 (1-\nu^2))`
        """
        return self.E * self.h**3 / (12 * (1 - self.nu**2))

    @property
    def D(self) -> float:
        """Alias for :class:`ElasticPlate.bending_stiffness`"""
        return self.bending_stiffness

    def omega0(self, a) -> float:
        return (self.bending_stiffness / (self.rho * a**2 * self.h)) ** 0.5


class _Simulation:
    def __init__(self, plate, res_array):
        self.plate = plate
        self.res_array = res_array

    @property
    def n_res(self):
        return len(self.res_array)

    def _strength(self, omega, resonator):
        # T_alpha in [Torrent2013]
        k = self.wavenumber(omega)
        # if isinstance(resonator, Pin):
        #     return 1j * 8 * k**2
        t_alpha = resonator.strength(omega) / self.plate.bending_stiffness
        return t_alpha / (1 - 1j * t_alpha / (8 * k**2))

    def _dstrength_domega(self, omega, resonator):
        k = self.wavenumber(omega)
        dk = self.dwavenumber_domega(omega)
        t_alpha = resonator.strength(omega) / self.plate.bending_stiffness
        dt_alpha_domega = (
            resonator.dstrength_domega(omega) / self.plate.bending_stiffness
        )
        Q = t_alpha / (8 * k**2)
        dQ = dt_alpha_domega / (8 * k**2)
        return 16 * k * (k / 2 * dQ - 1j * dk * Q**2) / (1 - 1j * Q) ** 2

    def wavenumber(self, omega):
        """Gives the wavenumber of the bare plate as a function of frequency `omega`

        Parameters
        ----------
        omega : float
            Frequency

        """
        return (
            omega**0.5
            * (self.plate.rho * self.plate.h / self.plate.bending_stiffness) ** 0.25
        )

    def dwavenumber_domega(self, omega):
        """Gives the derivative of the wavenumber with respect to the frequency `omega`

        Parameters
        ----------
        omega : float
            Frequency

        """
        k = self.wavenumber(omega)
        return (
            omega
            / (2 * k**3)
            * (self.plate.rho * self.plate.h / self.plate.bending_stiffness)
        )

    def plane_wave(self, x, y, omega, angle):
        """A plane wave

        Parameters
        ----------
        x : array
            x coordinates
        y : array
            y coordinates
        omega : float
            Frequency
        angle : float
            Angle in radians

        Returns
        -------
        array
            The plane wave field
        """
        k = self.wavenumber(omega)
        angle = bk.array(angle)
        return bk.exp(1j * k * (bk.cos(angle) * x + bk.sin(angle) * y))

    def point_source(self, x, y, omega, position):
        """A point source.

        Parameters
        ----------
        x : array
            x coordinates
        y : array
            y coordinates
        omega : float
            Frequency
        position : tuple
            Position of the source.


        Returns
        -------
        array
            The field from a line source
        """
        k = self.wavenumber(omega)
        xs, ys = position
        r = radial_coordinates(x - xs, y - ys)
        return greens_function(k, r)


class ScatteringSimulation(_Simulation):
    """Class to run a scattering simulation.

    Parameters
    ----------
    plate : :class:`ElasticPlate`
        The plate
    res_array : array of :class:`Pin`, :class:`Mass` or :class:`Resonator`
        An array containing the scatterers, it can be a mixture of pins, masses or resonators.
    alternative : bool
        Use an alternative formulation to solve for the scattering coefficients.

    """

    def __init__(self, plate, res_array, alternative=False):
        super().__init__(plate, res_array)

        self.alternative = alternative

        dists = bk.array(bk.zeros((self.n_res, self.n_res), dtype=bk.float64))
        for alpha, res_alpha in enumerate(self.res_array):
            xalpha, yalpha = res_alpha.position
            for beta, res_beta in enumerate(self.res_array):
                xbeta, ybeta = res_beta.position
                dists[alpha, beta] = radial_coordinates(xalpha - xbeta, yalpha - ybeta)

        self._dr = dists

    def build_rhs(self, phi0, omega, *args, **kwargs):
        phi0_vec = bk.array(bk.zeros((self.n_res), dtype=bk.complex128))
        for alpha, res_alpha in enumerate(self.res_array):
            xalpha, yalpha = res_alpha.position
            phi0_vec[alpha] = phi0(xalpha, yalpha, omega, *args, **kwargs)

        return phi0_vec

    def build_matrix(self, omega):
        omega = bk.array(omega)
        k = self.wavenumber(omega)
        matrix = bk.array(
            bk.zeros((self.n_res, self.n_res, *omega.shape), dtype=bk.complex128)
        )
        for alpha, res_alpha in enumerate(self.res_array):
            for beta, res_beta in enumerate(self.res_array):
                delta = 1 if alpha == beta else 0
                dr = self._dr[alpha, beta]
                G0 = greens_function(k, dr)
                if self.alternative:
                    t_beta = res_beta.strength(omega) / self.plate.D
                    mat = delta / t_beta - G0
                else:
                    T_beta = self._strength(omega, res_beta)
                    mat = delta - (1 - delta) * T_beta * G0
                matrix[alpha, beta] = mat
        return matrix

    def build_matrix_derivative(self, omega):
        omega = bk.array(omega)
        k = self.wavenumber(omega)
        matrix = bk.array(
            bk.zeros((self.n_res, self.n_res, *omega.shape), dtype=bk.complex128)
        )
        for alpha, res_alpha in enumerate(self.res_array):
            for beta, res_beta in enumerate(self.res_array):
                delta = 1 if alpha == beta else 0
                dr = self._dr[alpha, beta]
                dG0 = dgreens_function_dk(k, dr) * self.dwavenumber_domega(omega)

                if self.alternative:
                    t_beta = res_beta.strength(omega) / self.plate.D
                    dt_beta = res_beta.dstrength_domega(omega) / self.plate.D
                    mat = -dt_beta / t_beta**2 * delta - dG0
                else:
                    dT_beta = self._dstrength_domega(omega, res_beta)
                    G0 = greens_function(k, dr)
                    T_beta = self._strength(omega, res_beta)
                    mat = -(1 - delta) * (dT_beta * G0 + T_beta * dG0)
                matrix[alpha, beta] = mat
        return matrix

    # def get_eigenvector(self, omega):
    #     M = self.build_matrix(omega)
    #     return null(M)

    def solve(self, incident, omega, *args, **kwargs):
        """Solve the multiple scattering problem

        Parameters
        ----------
        incident : callable
            The incident field. Signature should be ``incident(x, y, omega, *args, **kwargs)``.
        omega: float
            Frequency
        Returns
        -------
        array
            The solution vector
        """
        matrix = self.build_matrix(omega)
        rhs = self.build_rhs(incident, omega, *args, **kwargs)
        return bk.linalg.solve(matrix, rhs)

    def _get_field(
        self, x, y, incident, kernel, solution_vector, omega, *args, **kwargs
    ):
        k = self.wavenumber(omega)
        if incident is None:
            W = 0
        else:
            W = incident(x, y, omega, *args, **kwargs)
        for alpha, res_alpha in enumerate(self.res_array):
            xalpha, yalpha = res_alpha.position
            G0 = kernel(k, x - xalpha, y - yalpha)
            coeff = 1 if self.alternative else self._strength(omega, res_alpha)
            W += coeff * solution_vector[alpha] * G0
        return W

    def get_field(self, x, y, incident, solution_vector, omega, *args, **kwargs):
        """Get the total field.

        Parameters
        ----------
        x : array
            x coordinates
        y : array
            y coordinates
        incident : callable
            The incident field. Signature should be ``incident(x, y, omega, *args, **kwargs)``.
        solution_vector: array
            The solution vector as returned by :class:`ScatteringSimulation.solve`
        omega: float or complex
            Frequency

        Returns
        -------
        array
            The total field
        """
        return self._get_field(
            x,
            y,
            incident,
            greens_function_cartesian,
            solution_vector,
            omega,
            *args,
            **kwargs
        )

    def get_scattered_field(self, x, y, solution_vector, omega):
        """Get the scattered field.

        Parameters
        ----------
        x : array
            x coordinates
        y : array
            y coordinates
        solution_vector: array
            The solution vector as returned by :class:`ScatteringSimulation.solve`
        omega: float or complex
            Frequency


        Returns
        -------
        array
            The scattered field
        """
        return self._get_field(
            x, y, None, greens_function_cartesian, solution_vector, omega
        )

    def eigensolve(self, *args, **kwargs):
        return nonlinear_eigensolver(self, *args, **kwargs)

    def get_mode(self, x, y, eigenvector, eigenvalue):
        return self.get_scattered_field(x, y, eigenvector, eigenvalue)

    def get_strength_poles(self):
        singularity_Talpha = []
        plate = self.plate
        for res in self.res_array:
            q = (
                1
                / 16
                * (res.mass * res.stiffness / (plate.rho * plate.h * plate.D)) ** 0.5
                + 0j
            )
            pole = res.omega_r * (-1j * q + (1 - q**2 + 0j) ** 0.5)
            singularity_Talpha.append(pole)
            pole = res.omega_r * (-1j * q - (1 - q**2) ** 0.5)
            singularity_Talpha.append(pole)
        return bk.array(singularity_Talpha)

    # FAR FIELD QUANTITIES

    def get_far_field_radiation(self, phi_n, omega, angle):
        """
        Compute the far field radiation pattern.

        Equation (12) in :cite:p:`marti2023bound`

        Parameters
        ----------
        phi_n : array
            The scattering coefficients at the position of the scatterers by using
            the second definition of the M matrix
        omega : float
            The frequency
        angle: array
            Angle(s) in radian

        Returns
        -------
        F : array
             The far field radiation pattern
        """

        k = self.wavenumber(omega)

        F = bk.zeros(len(angle), dtype=bk.complex64)
        for thetaIterator, theta in enumerate(angle):
            for alpha, res_alpha in enumerate(self.res_array):
                xalpha, yalpha = res_alpha.position
                angleScatterer = bk.arctan2(yalpha, xalpha)
                moduleDistance = bk.sqrt(xalpha**2 + yalpha**2)
                F[thetaIterator] += phi_n[alpha] * bk.exp(
                    -1j * k * moduleDistance * bk.cos(theta - angleScatterer)
                )
        return F

    def get_scattering_cross_section(self, phi_n, omega):
        """Compute the scattering cross section of the cluster in the farfield.

        Equation (A6b) in :cite:p:`packo2021metaclusters`

        Parameters
        ----------
        phi_n : array
            The scattering coefficients at the position of the scatterers by using
            the second definition of the M matrix#
        omega : float
            The frequency

        Returns
        -------
        sigma_sc : float
             Scattering cross section
        """
        k = self.wavenumber(omega)
        sigma_sc = (
            1 / (8 * self.plate.bending_stiffness * k**2) * bk.sum(bk.abs(phi_n) ** 2)
        )
        return sigma_sc

    def get_absorption_cross_section(self, phi_n, omega, sum=True):
        """Compute the absorption cross section of the cluster in the farfield.

        Equation (A6c) in :cite:p:`packo2021metaclusters`

        Parameters
        ----------
        phi_n : array
            The scattering coefficients at the position of the scatterers by using
            the second definition of the M matrix#
        omega : float
            The frequency
        sum: bool
            Wether to return the sum of all resonator contributions

        Returns
        -------
        sigma_abs : float or array
             Absorption cross section (total if sum=True, else array with value for each resonator)
        """
        sigma_abs = bk.zeros(self.n_res)
        for resIterator, res in enumerate(self.res_array):
            if self.alternative == False:
                acs = bk.imag(res.strength(omega)) * bk.abs(phi_n) ** 2
            else:
                acs = (
                    bk.imag(res.strength(omega))
                    * bk.abs(phi_n / self._strength(omega, res)) ** 2
                )
            sigma_abs[resIterator] = acs

        if sum:
            return bk.sum(sigma_abs)
        return sigma_abs

    def get_extintion_cross_section(self, phi_n, omega):
        """Compute the extintion cross section of the cluster in the farfield

        Equation (A6a) in :cite:p:`packo2021metaclusters`

        Parameters
        ----------
        phi_n : array
            The scattering coefficients at the position of the scatterers by using
            the second definition of the M matrix#
        omega : float
            The frequency

        Returns
        -------
        sigma_ext : float
             Extintion cross section
        """
        k = self.wavenumber(omega)
        F = self.get_far_field_radiation(phi_n, omega, [0])
        return bk.imag(F[0])

    def get_energy_efficiency(self, phi_n, omega):
        """
        Equation (17) in :cite:p:`packo2021metaclusters`
        """
        sigma_sc = self.get_scattering_cross_section(self, phi_n, omega)
        sigma_abs = self.get_extintion_cross_section(self, phi_n, omega)
        return sigma_sc / sigma_abs

    def get_multiplicities(self, eigenvalues, M=None, tol=1e-6):
        if M is None:
            M = [self.build_matrix(omega) for omega in eigenvalues]
        multiplicities = [self.n_res - bk.linalg.matrix_rank(m, tol=tol) for m in M]
        return multiplicities

    def update_eig(self, evs, eigenvectors, multiplicities):
        Nmodes = len(evs)
        eigvals = []
        eigmodes = []
        for i in range(Nmodes):
            mult = multiplicities[i]
            if mult > 1:
                # vs = kl.eig.null(M[i], mult=mult)
                v = eigenvectors[:, i]
                # _A = [v for l in range(mult)]
                # A = bk.vstack(_A).T
                # vs = kl.eig.gram_schmidt(A)
                # vs = [_.flatten() for _ in vs]
                vs = bk.vstack([v] * mult)
                for jm in range(mult):
                    eigvals.append(evs[i])
                    # eigmodes.append(vs[jm])
                    eigmodes.append(bk.roll(vs[jm], 2 * jm))
            else:
                eigvals.append(evs[i])
                eigmodes.append(eigenvectors[:, i])

        eigenvectors = bk.stack(eigmodes).T
        evs = bk.array(eigvals)
        return evs, eigenvectors


class BandsSimulation(_Simulation):
    """Class to compute phononic bands.

    Parameters
    ----------
    plate : :class:`ElasticPlate`
        The plate
    res_array : list
        Array of scatterers. This can be :class:`Pin`, :class:`Mass` or :class:`Resonator` or a mix of those.
    lattice_vectors : tuple of tuples
        The two vectors `(v1x, v1y), (v2x, v2y)` defining the lattice.

    """

    def __init__(self, plate, res_array, lattice_vectors):
        super().__init__(plate, res_array)

        self.lattice_vectors = bk.array(
            lattice_vectors, dtype=bk.float64
        )  #: The lattice vectors
        self.reciprocal = 2 * bk.pi * bk.linalg.inv(self.lattice_vectors).T

        self.a = min([bk.linalg.norm(v) for v in self.lattice_vectors])
        self.omega0 = plate.omega0(self.a)
        v0, v1 = self.lattice_vectors
        self.area = bk.abs(v0[0] * v1[1] - v0[1] * v1[0])

    def _reciprocal_sum(self, Omega, K, alpha, beta, M):
        out = 0
        for m in range(-M, M + 1):
            for n in range(-M, M + 1):
                G = m * self.reciprocal[0] + n * self.reciprocal[1]
                Rab = bk.array(self.res_array[beta].position) - bk.array(
                    self.res_array[alpha].position
                )
                out += bk.exp(-1j * G @ Rab) / (
                    bk.linalg.norm(K + G) ** 4 * self.a**4 - Omega**2 * self.a**2
                )
        return out

    def _gamma(self, beta):
        return self.res_array[beta].mass / (self.plate.rho * self.plate.h * self.area)

    def _matrix_element(self, Omega, K, alpha, beta, M):
        S = self._reciprocal_sum(Omega, K, alpha, beta, M)
        Omega_beta = self.res_array[beta].omega_r / self.omega0
        return (
            self._gamma(beta)
            * Omega**2
            * self.a**2
            / (1 - Omega**2 / Omega_beta**2)
        ) * S

    def _build_matrix(self, Omega, K, M):
        matrix = bk.array(bk.zeros((self.n_res, self.n_res), dtype=bk.complex128))
        for alpha in range(self.n_res):
            for beta in range(self.n_res):
                matrix[alpha, beta] = self._matrix_element(Omega, K, alpha, beta, M)
        return bk.eye(self.n_res) - matrix

    def _getP(self, alpha, M):
        P = []
        for m in range(-M, M + 1):
            for n in range(-M, M + 1):
                G = bk.array(
                    m * self.reciprocal[0] + n * self.reciprocal[1], dtype=bk.float64
                )
                Ra = bk.array(self.res_array[alpha].position, dtype=bk.float64)
                ph = bk.exp(1j * (Ra @ G))
                P.append(ph)
        return bk.array(P) if P == [] else bk.stack(P)

    def _getK(self, K, M):
        K = bk.array(K)
        P = []
        for m in range(-M, M + 1):
            for n in range(-M, M + 1):
                G = bk.array(
                    m * self.reciprocal[0] + n * self.reciprocal[1], dtype=bk.float64
                )
                P.append(bk.linalg.norm(K + G) ** 4)
        return bk.diag(bk.array(P, dtype=bk.float64))

    def eigensolve(self, k, Npw, return_modes=True, hermitian=False):
        """Solve the eigenvalue problem.

        Parameters
        ----------
        k : tuple
            Wavevector
        Npw : int
            Number of plane waves in each direction (from `-Npw` to `Npw`)
        return_modes : bool, optional
            Compute eigenmodes?
        hermitian : bool, optional
            Is the problem Hermitian?

        Returns
        -------
        array or tuple of arrays
            Eigenvalues (and eigenmodes if `return_modes` is `True`)
        """
        N = (2 * Npw + 1) ** 2

        Ps = []
        for alpha in range(self.n_res):
            P = self._getP(alpha, Npw)
            Ps.append(P)
        Ps = bk.array(Ps) if Ps == [] else bk.stack(Ps)

        Kmat0 = self._getK(k, Npw)

        Nmat = self.n_res + N
        Kmat = bk.zeros((Nmat, Nmat), dtype=bk.complex128)
        Q = 0
        stiffnesses = []
        masses = []
        gamma_inf = _BIG
        for alpha, res in enumerate(self.res_array):
            Omega_inf_square = _BIG if isinstance(res, Mass) else 1
            mass = (
                self.plate.rho * self.plate.h * self.area * gamma_inf
                if isinstance(res, Pin)
                else res.mass
            )
            stiffness = (
                res.stiffness
                if isinstance(res, Resonator)
                else Omega_inf_square
                / self.plate.D
                * mass
                / (self.plate.rho * self.plate.h)
            )

            stiffnesses.append(stiffness)
            masses.append(mass)
            P = bk.stack([Ps[alpha]]).T
            Q += P @ bk.conj(P).T * stiffness
            Kmat[N + alpha, :N] = -stiffness * bk.conj(Ps[alpha])
            Kmat[:N, N + alpha] = -stiffness * (Ps[alpha])

        Kmat[:N, :N] = self.plate.bending_stiffness * self.area * Kmat0 + Q
        Kmat[N:, N:] = bk.diag(bk.array(stiffnesses, dtype=bk.float64))

        Mmat = bk.zeros((Nmat), dtype=bk.complex128)
        Mmat[:N] = self.plate.rho * self.plate.h * self.area
        Mmat[N:] = bk.array(masses)

        self.Kmat = Kmat
        self.Mmat = Mmat
        Amat = Kmat / Mmat

        if return_modes:
            eigensolve = bk.linalg.eigh if hermitian else bk.linalg.eig
            eigvals, modes = eigensolve(Amat)
        else:
            eigensolve = bk.linalg.eigvalsh if hermitian else bk.linalg.eigvals
            eigvals = eigensolve(Amat)
        omegans = eigvals**0.5
        isort = bk.argsort(omegans.real)
        if return_modes:
            return omegans[isort], modes[:, isort]
        return omegans[isort]

    def get_field(self, mode_coeffs, x, y):
        """Get the mode field

        Parameters
        ----------
        mode_coeffs : array
            Fourrier coefficients of the mode
        x : array
            x coordinates
        y : array
            y coordinates

        Returns
        -------
        array
            Modal field in real space
        """
        M = int(((len(mode_coeffs) - self.n_res) ** 0.5 - 1) / 2)
        field = 0
        i = 0
        for m in range(-M, M + 1):
            for n in range(-M, M + 1):
                G = m * self.reciprocal[0] + n * self.reciprocal[1]
                kr = G[0] * x + G[1] * y
                field += mode_coeffs[i] * bk.exp(-1j * kr)
                i += 1
        return field


class DiffractionSimulation(_Simulation):
    """Class to run a diffraction simulation.

    Parameters
    ----------
    plate : :class:`ElasticPlate`
        The plate
    res_array : array of :class:`Pin`, :class:`Mass` or :class:`Resonator`
        An array containing the scatterers, it can be a mixture of pins, masses or resonators.
    period : float
        Period of the grating
    nh : int
        Number of harmonics such that the calculations use orders [-nh,...0, ...,nh]

    """

    def __init__(self, plate, res_array, period, nh=0):
        super().__init__(plate, res_array)

        for res in self.res_array:
            if not (0 < res.position[0] < period):
                raise ValueError("Resonator x position must be between 0 and period")

        self.period = period
        self.nh = nh

    @property
    def harmonics(self):
        return bk.array(range(-self.nh, self.nh + 1))

    def _G0(self, k):
        return 1 / (4 * self.period * k**2)

    def _g(self, n):
        return 2 * n * bk.pi / self.period

    def _get_xis(self, k, kx, n):
        qn = kx + self._g(n)
        xip = (qn**2 + k**2 + 0j) ** 0.5
        xim = (qn**2 - k**2 + 0j) ** 0.5
        return xip.conj(), xim.conj()

    def is_propagative(self, k, kx, n):
        return bk.abs(kx + self._g(n)) < k

    def _lattice_sum(self, k, kx, xa, ya, xb, yb, harmonics=None):
        harmonics = harmonics or self.harmonics

        dx = xa - xb
        abs_dy = bk.abs(ya - yb)
        lat_sum = 0
        for n in harmonics:
            xip, xim = self._get_xis(k, kx, n)
            t = bk.exp(-xim * abs_dy) / xim
            t -= bk.exp(-xip * abs_dy) / xip
            t *= bk.exp(1j * (kx + self._g(n)) * dx)
            lat_sum += t
        return lat_sum

    def build_matrix(self, omega, kx):
        omega = bk.array(omega)
        k = self.wavenumber(omega)
        G0 = self._G0(k)
        matrix = bk.array(
            bk.zeros((self.n_res, self.n_res, *omega.shape), dtype=bk.complex128)
        )
        for alpha, res_alpha in enumerate(self.res_array):
            for beta, res_beta in enumerate(self.res_array):
                delta = 1 if alpha == beta else 0
                t_beta = res_beta.strength(omega) / self.plate.D
                chi = G0 * self._lattice_sum(
                    k, kx, *res_alpha.position, *res_beta.position
                )
                matrix[alpha, beta] = delta / t_beta - chi
        return matrix

    def build_rhs(self, omega, angle):
        phi0_vec = bk.array(bk.zeros((self.n_res), dtype=bk.complex128))
        for alpha, res_alpha in enumerate(self.res_array):
            phi0_vec[alpha] = self.plane_wave(*res_alpha.position, omega, angle)

        return phi0_vec

    def solve(self, omega, angle):
        """Solve the diffraction problem

        Parameters
        ----------
        omega: float
            Frequency
        angle: float
            Incident angle (in radians)
        Returns
        -------
        array
            The solution vector
        """
        k = self.wavenumber(omega)
        angle = bk.array(angle)
        kx = k * bk.cos(angle)
        matrix = self.build_matrix(omega, kx)
        rhs = self.build_rhs(omega, angle)
        return bk.linalg.solve(matrix, rhs)

    def _get_field(self, x, y, incident, solution_vector, omega, angle):
        k = self.wavenumber(omega)
        angle = bk.array(angle)
        kx = k * bk.cos(angle)
        G0 = self._G0(k)
        if incident is None:
            W = 0
        else:
            W = incident(x, y, omega, angle)
        for alpha, res_alpha in enumerate(self.res_array):
            chi = G0 * self._lattice_sum(k, kx, x, y, *res_alpha.position)
            W += solution_vector[alpha] * chi
        return W

    def get_field(self, x, y, solution_vector, omega, angle):
        """Get the total field.

        Parameters
        ----------
        x : array
            x coordinates
        y : array
            y coordinates
        solution_vector: array
            The solution vector as returned by :class:`ScatteringSimulation.solve`
        omega: float or complex
            Frequency
        angle: float
            Incident angle (in radians)

        Returns
        -------
        array
            The total field
        """
        return self._get_field(
            x,
            y,
            self.plane_wave,
            solution_vector,
            omega,
            angle,
        )

    def get_scattered_field(self, x, y, solution_vector, omega, angle):
        """Get the scattered field.

        Parameters
        ----------
        x : array
            x coordinates
        y : array
            y coordinates
        solution_vector: array
            The solution vector as returned by :class:`ScatteringSimulation.solve`
        omega: float or complex
            Frequency
        angle: float
            Incident angle (in radians)


        Returns
        -------
        array
            The scattered field
        """
        return self._get_field(
            x,
            y,
            None,
            solution_vector,
            omega,
            angle,
        )

    def orders_angle(self, k, angle, n):
        angle = bk.array(angle)
        return bk.arccos(bk.cos(angle) + 2 * n * bk.pi / (self.period * k))

    @staticmethod
    def _wavevector_orders(k, thetan):
        thetan = bk.array(thetan)
        kx, ky = k * bk.cos(thetan), k * bk.sin(thetan)
        return bk.array([kx, ky]), bk.array([kx, -ky])

    def get_propagative_harmonics(self):
        propa = [self.is_propagative(k, kx, n) for n in self.harmonics]
        propagative = []
        for i, prop in enumerate(propa):
            if prop:
                propagative.append(self.harmonics[i])
        return propagative

    def get_efficiencies(self, sol, omega, angle):
        angle = bk.array(angle)
        k = self.wavenumber(omega)
        kx = k * bk.cos(angle)
        G0 = self._G0(k)

        propa = [self.is_propagative(k, kx, n) for n in self.harmonics]
        t, r, T, R = {}, {}, {}, {}
        sumT = sumR = 0

        for n in self.harmonics:
            if self.is_propagative(k, kx, n):
                thetan = self.orders_angle(k, angle, n)
                cst = 1j * G0 / (k * bk.sin(thetan))
                knp, knm = self._wavevector_orders(k, thetan)
                rn = tn = 0
                for beta, res in enumerate(self.res_array):
                    pos = bk.array(res.position)
                    tn += sol[beta] * bk.exp(-1j * (knp @ pos))
                    rn += sol[beta] * bk.exp(-1j * (knm @ pos))
                rn *= cst
                tn *= cst
                if n == 0:
                    tn += 1

                nrm_ = bk.sin(thetan) / bk.sin(angle)
                Rn = bk.abs(rn) ** 2 * nrm_
                Tn = bk.abs(tn) ** 2 * nrm_

                sumR += Rn
                sumT += Tn

            else:
                rn = tn = Rn = Tn = 0

            key = str(int(n))
            r[key] = rn
            t[key] = tn
            R[key] = Rn
            T[key] = Tn
        Rdict = dict(amplitude=r, energy=R, total=sumR)
        Tdict = dict(amplitude=t, energy=T, total=sumT)
        return Rdict, Tdict

    def get_min_harmonics(self, omega_max, angle, nhmax=100):
        k = self.wavenumber(omega_max)
        angle = bk.array(angle)
        kx = k * bk.cos(angle)
        nh_ = 0
        propa = True

        while propa:
            nh_ += 1
            propa1 = self.is_propagative(k, kx, -nh_)
            propa2 = self.is_propagative(k, kx, nh_)
            propa = propa1 or propa2
            if nh_ > nhmax:
                break
        nh_ -= 1

        return nh_
